import sys
import traceback
import socket
import socketserver
import http.server

from acmems import exceptions


class ThreadedACMEServerInet4(socketserver.ThreadingMixIn,
                              http.server.HTTPServer):
    daemon_threads = True
    allow_reuse_address = True


class ThreadedACMEServerInet6(socketserver.ThreadingMixIn,
                              http.server.HTTPServer):
    daemon_threads = True
    allow_reuse_address = True
    address_family = socket.AF_INET6


ThreadedACMEServerByType = {
    socket.AF_INET: ThreadedACMEServerInet4,
    socket.AF_INET6: ThreadedACMEServerInet6,
}


class ACMEAbstractHandler(http.server.BaseHTTPRequestHandler):
    server_version = 'AcmeManager/0.1'
    manager = None

    def send_data(self, data, content_type='text/plain', response_code=200):
        """ Helper method to send data as HTTP response. The data are
            transfered as :mimetype:`text/plain`.

            :param str data: The text to send as :py:obj:`Python String <str>`.
            :param int response_code: HTTP response code"""
        if type(data) is not bytes:
            data = str(data).encode('utf-8')
        self.send_response(response_code)
        self.send_header('Content-type', content_type)
        self.send_header('Content-Length', str(len(data)))
        self.end_headers()
        self.wfile.write(data)


class ACMEHTTPHandler(ACMEAbstractHandler):
    def __init__(self, validator, *args, **kwargs):
        self.validator = validator
        super().__init__(*args, **kwargs)

    def do_GET(self):
        """ Handles POST request (upload files). """
        host = self.headers['Host']
        if host.endswith(':5002'):
            host = host[:-5]
        try:
            self.send_data(self.validator.response_for(host, self.path))
        except KeyError:
            self.send_error(404)


class ACMEMgmtHandler(ACMEAbstractHandler):
    @property
    def auth(self):
        return self.manager.config.auth

    def do_POST(self):
        """ Handles POST request (upload files). """
        if self.path != '/sign':
            self.send_error(404)
            return
        try:
            with self.auth.process(self.client_address, self.headers, self.rfile) as p:
                if not p.acceptable():
                    self.send_error(403)
                    return
                print(self.client_address, p.common_name, p.dns_names)
                cached_certs = p.storage.from_cache(p.csrpem)
                if cached_certs:
                    self.send_data(cached_certs)
                    return
                authzrs = self.manager.acquire_domain_validations(p.validator, p.dns_names)
                certs = '\n'.join(self.manager.issue_certificate(p.csr, authzrs))
                print(certs)
                p.storage.add_to_cache(p.csrpem, certs)
                self.send_data(certs)
        except exceptions.PayloadToLarge:
            self.send_error(413)
        except exceptions.PayloadInvalid:
            self.send_error(415)
        except Exception:
            self.send_error(500)
            print('-'*50, file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
            print('-'*50, file=sys.stderr)
            sys.stderr.flush()



import os
import os.path
from hashlib import sha384
from datetime import datetime, timedelta

from cryptography import x509
from cryptography.hazmat.backends import default_backend

from acmems.config import ConfigurationError


class StorageImplementor():
    def __init__(self, type, name, options):
        self.type = type
        self.name = name
        self.parse(options)


class NoneStorageImplementor(StorageImplementor):
    def parse(self, options):
        if len(options) > 0:
            raise ConfigurationError('none storage does not support any options, but found "{}"'.format('", "'.join(o[0] for o in options)))

    def from_cache(self, csr):
        return None

    def add_to_cache(self, csr, certs):
        return None


class FileStorageImplementor(StorageImplementor):
    def parse(self, options):
        self.directory = None
        self.renew_within = None
        for option, value in options:
            if option == 'directory':
                self.directory = value
            elif option == 'renew-within':
                self.renew_within = timedelta(days=int(value))
            else:
                raise ConfigurationError('FileStorage: unknown option "{}"'.format(option))
        if self.directory is None:
            raise ConfigurationError('FileStorage: option directory is required')
        if self.renew_within is None:
            self.renew_within = timedelta(days=14)

    def cache_dir(self, csr):
        hash = sha384(csr).hexdigest()
        return os.path.join(self.directory, hash[0:2], hash[2:])

    def from_cache(self, csr):
        dir = self.cache_dir(csr)
        if not os.path.isfile(os.path.join(dir, 'csr.pem')):
            return None
        if not os.path.isfile(os.path.join(dir, 'cert.pem')):
            return None
        if csr != open(os.path.join(dir, 'csr.pem'), 'rb').read():
            # should not happen!!
            return None
        certpem = open(os.path.join(dir, 'cert.pem'), 'rb').read()
        cert = x509.load_pem_x509_certificate(certpem, default_backend())
        current_validation_time = cert.not_valid_after - datetime.now()
        if current_validation_time < self.renew_within:
            return None
        else:
            return certpem.decode('utf-8')

    def add_to_cache(self, csr, cert):
        dir = self.cache_dir(csr)
        os.makedirs(dir, exist_ok=True)
        with open(os.path.join(dir, 'csr.pem'), 'bw') as f:
            f.write(csr)
        with open(os.path.join(dir, 'cert.pem'), 'w') as f:
            f.write(cert)
        return True


implementors = {
    'none': NoneStorageImplementor,
    'file': FileStorageImplementor,
}


def setup(type, name, options):
    try:
        return implementors[type](type, name, options)
    except KeyError:
        raise ConfigurationError('Unsupported storage type "{}"'.format(type))



import os.path
import socket
import warnings
import re

from acmems.auth import Authenticator


class ConfigurationError(Exception):
    pass


class MissingSectionError(ConfigurationError):
    pass


class UnknownVerificationError(ConfigurationError):
    pass


class UnknownStorageError(ConfigurationError):
    pass


class SingletonOptionRedifined(ConfigurationError):
    def __init__(self, section, option, old, new):
        self.section = section
        self.option = option
        self.old = old
        self.new = new

    def __str__(self):
        return 'Singleton option redefined: {}.{} was {}, redefined as {}'.format(self.section, self.option, self.old, self.new)


class ConfigurationWarning(UserWarning):
    pass


class UnusedOptionWarning(ConfigurationWarning):
    pass


class OptionRedifinitionWarning(ConfigurationWarning):
    pass


class UnusedSectionWarning(ConfigurationWarning):
    pass


class Configurator():
    def __init__(self, *configs):
        self.validators = {}
        self.default_validator = None
        self.storages = {}
        self.default_storage = None
        self._max_size = None

        self.auth = Authenticator(self)
        for config in configs:
            self.parse(config)

    @property
    def keyfile(self):
        return os.path.join(self.account_dir, 'account.pem')

    @property
    def registration_file(self):
        return os.path.join(self.account_dir, 'registration.json')

    @property
    def max_size(self):
        return self._max_size or 4096

    def parse(self, config):
        config = self.read_data(config)
        self.parse_account_config(config.pop('account'))
        self.parser_mgmt_config(config.pop('mgmt'))
        special_group_re = re.compile('^(?P<type>(auth|verification|storage)) (?P<opener>"?)(?P<name>.+)(?P=opener)$')
        auth_blocks = []
        for group, options in config.items():
            match = special_group_re.match(group)
            if match:
                if match.group('type') == 'auth':
                    # parse auth blocks last to have verification and storage blocks processed
                    auth_blocks.append((match.group('name'), options))
                elif match.group('type') == 'verification':
                    self.parse_verification_group(match.group('name'), options)
                else:
                    self.parse_storage_group(match.group('name'), options)
            else:
                warnings.warn('Unknown section name: {0}'.format(group),
                              UnusedSectionWarning, stacklevel=2)

        self.setup_default_validator()
        self.setup_default_storage()

        for name, options in auth_blocks:
            self.auth.parse_block(name, options)

    @staticmethod
    def read_data(config):
        """ Reads the given file name. It assumes that the file has a INI file
            syntax. The parser returns the data without comments and fill
            characters. It supports multiple option with the same name per
            section but not multiple sections with the same name.

            :param str filename: path to INI file
            :return: a dictionary - the key is the section name value, the
                option is a array of (option name, option value) tuples"""
        sections = {}
        with config as f:
            section = None
            options = None
            for line in f:
                line = line.strip()
                # ignore comments:
                if line.startswith('#') or line.startswith(';'):
                    continue
                if not line:
                    continue
                # handle section header:
                if line.startswith('[') and line.endswith(']'):
                    if section:  # save old section data
                        sections[section] = options
                    section = line[1:-1]
                    options = []
                    continue
                if section is None:
                    warnings.warn('Option without sections: {0}'.format(line),
                                  UnusedOptionWarning, stacklevel=2)
                    continue
                option, value = line.split('=', 1)
                options.append((option.strip(), value.strip()))
            if section:  # save old section data
                sections[section] = options
        return sections

    def parse_account_config(self, config):
        self.acme_server = None
        for option, value in config:
            if option == 'acme-server':
                if self.acme_server is not None:
                    raise SingletonOptionRedifined(
                        section='account',
                        option='acme_server',
                        old=self.acme_server,
                        new=value)
                self.acme_server = value
            elif option == 'dir':
                self.account_dir = value
            else:
                warnings.warn('Option unknown [{}]{} = {}'.format('account', option, value),
                              UnusedOptionWarning, stacklevel=2)
        if self.acme_server is None:
            self.acme_server = 'https://acme-staging.api.letsencrypt.org/directory'

    def parser_mgmt_config(self, config):
        self.mgmt_listeners = None
        for option, value in config:
            if option == 'max-size':
                suffixes = {'k': 1024, 'm': 1024*1024}
                for suffix, mul in suffixes.items():
                    if value.endswith(suffix):
                        self._max_size = int(value[:len(suffix)]) * mul
                        break
                else:
                    self._max_size = int(value)
            elif option == 'default-verification':
                if value == '':
                    self.default_validator = False
                else:
                    self.default_validator = value.strip()
            elif option == 'default-storage':
                if value == '':
                    self.default_storage = False
                else:
                    self.default_storage = value.strip()
            elif option == 'listener':
                if self.mgmt_listeners is None:
                    self.mgmt_listeners = []
                if value == '':  # disable listener
                    continue
                if ':' not in value:
                    raise ConfigurationError('unix socket are currenlty not supported as listeners')
                host, port = value.rsplit(':', 1)
                if host[0] == '[' and host[-1] == ']':
                    host = host[1:-1]
                self.mgmt_listeners += socket.getaddrinfo(host, int(port), proto=socket.IPPROTO_TCP)
            else:
                warnings.warn('Option unknown [{}]{} = {}'.format('listeners', option, value),
                              UnusedOptionWarning, stacklevel=2)
        if self.mgmt_listeners is None:
            self.mgmt_listeners = socket.getaddrinfo('127.0.0.1', 1313, proto=socket.IPPROTO_TCP) \
                + socket.getaddrinfo('::1', 1313, proto=socket.IPPROTO_TCP)

    def parse_verification_group(self, name, options):
        option, value = options.pop(0)
        if option != 'type':
            raise ConfigurationError('A verification must start with the type value!')
        from acmems.challenges import setup
        self.validators[name] = setup(value, name, options)

    def setup_default_validator(self):
        if self.default_validator is False:  # default validator disabled
            return
        if self.default_validator:  # defined
            self.default_validator = self.validators[self.default_validator]
            return
        if len(self.validators) == 1:  # we use the only defined validator as default
            self.default_validator = list(self.validators.values())[0]
        else:  # define a default http storage
            from acmems.challenges import setup
            self.default_validator = self.validators['http'] = setup('http01', 'http', ())

    def parse_storage_group(self, name, options):
        option, value = options.pop(0)
        if option != 'type':
            raise ConfigurationError('A storage must start with the type value!')
        from acmems.storages import setup
        self.storages[name] = setup(value, name, options)

    def setup_default_storage(self):
        if self.default_storage is False:  # default storage disabled
            return
        if self.default_storage:
            self.default_storage = self.storages[self.default_storage]
        if len(self.storages) == 1:
            self.default_storage = list(self.storages.values())[0]
        else:
            from acmems.storages import setup
            self.default_storage = self.storages['none'] = setup('none', 'none', ())



import os.path
import time
from datetime import datetime, timedelta

import OpenSSL.crypto

import acme.client
import acme.messages
import acme.jose
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization as pem_serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from acmems import exceptions


class ACMEManager():
    ''' ACME manager - high level ACME client; process authorizations via
        http01 automatically.

    :ivar dict responses: Responses to deliver; designed as answers for
        authorization challenges. dict[host][path] = value
    :ivar dict authzrs: List of current active `acme.messages.AuthorizationResource`
    :ivar acmems.config.Configuration config: Active configuration

    '''
    def __init__(self, config, connect=True):
        self.responses = {}
        self.authzrs = {}
        self.config = config
        if connect:
            self.connect()

    def log(self, *args):
        ''' log something

        .. todo::
            Switch to real logging'''
        print(*args)

    # ----------------------------------------------------------
    # 1. generall ACME account handling (keys, registration ...)
    # ----------------------------------------------------------

    def connect(self):
        ''' initialize/setup ourself; load private key, create ACME client
            and refresh our registration

            :raises acmems.exceptions.AccountError: could not load account
            :raises acmems.exceptions.NeedToAgreeToTOS: terms of service are
                not accepted - cannot operate
        '''
        self.load_private_key()
        self.init_client()
        self.refresh_registration()

    def load_private_key(self):
        ''' load our private key / the key to identify ourself against
            the ACME server. This key MUST NOT be used for certificates.

            :raises acmems.exceptions.AccountError: something is broken
                with our account (mustly key not found)
        '''
        try:
            with open(self.config.keyfile, 'rb') as f:
                key = pem_serialization.load_pem_private_key(
                    f.read(),
                    password=None,
                    backend=default_backend())
        except FileNotFoundError:
            raise exceptions.AccountError('Key {} not found'
                                          .format(self.config.keyfile))
        # TODO - handle IOError; keyfile without valid key
        self.key = acme.jose.JWKRSA(key=acme.jose.ComparableRSAKey(key))

    def create_private_key(self, force=False, key_size=4096):
        ''' create new private key to be used for identify ourself against
            the ACME server

            Key is afterwards read via `load_private_key`!

            :param bool force: create new key even key exists already
            :param int key_size: private key size in bits (at least 2048)

            :raises acmems.exceptions.AccountError: account dir not found
                or private key will not be overriden (force is `False`).
        '''
        if os.path.isfile(self.config.keyfile) and not force:
            raise exceptions.AccountError('Existing key is only override'
                                          ' if I am forced to')
        key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=key_size,
            backend=default_backend())
        with open(self.config.keyfile, 'wb') as f:
            f.write(key.private_bytes(
                encoding=pem_serialization.Encoding.PEM,
                format=pem_serialization.PrivateFormat.PKCS8,
                encryption_algorithm=pem_serialization.NoEncryption(),
            ))
        # verify that everythings works by reading the key from disk
        self.load_private_key()

    def init_client(self):
        ''' create ACME client
        '''
        self.client = acme.client.Client(self.config.acme_server, self.key)

    def register(self, emails=[]):
        resource = acme.messages.NewRegistration(
            key=self.key.public_key(),
            contact=tuple(
                ['mailto:{}'.format(mail) for mail in emails]))
        self.registration = self.client.register(resource)
        self.dump_registration()

    def tos_agreement_required(self):
        if self.registration.body.agreement is not True:
            return self.registration.terms_of_service

    def accept_terms_of_service(self, url):
        newreg = self.registration.update(
            body=self.registration.body.update(agreement=url)
        )
        self.registration = self.client.update_registration(newreg)

    def dump_registration(self):
        with open(self.config.registration_file, 'w') as f:
            f.write(self.registration.json_dumps_pretty())

    def refresh_registration(self):
        # Register or validate and update our registration.
        existing_regr = None

        # Validate existing registration by querying for it from the server.
        try:
            with open(self.config.registration_file, 'r') as f:
                self.registration = acme.messages.RegistrationResource \
                    .json_loads(f.read())
        except FileNotFoundError:
            raise exceptions.AccountError('Key is not yet registered'
                                          ' or registration is losted!')
        existing_regr = self.registration.json_dumps()
        self.registration = self.client.query_registration(self.registration)

        if existing_regr != self.registration.json_dumps():
            self.dump_registration()

        # the terms of server needs to be agreed to use the ACME server!
        if not self.registration.body.agreement:
            raise exceptions.NeedToAgreeToTOS(
                self.registration.terms_of_service)

    # ---------------------------------------------------------
    # 2. the real part (handling authorizations + certificates)
    # ---------------------------------------------------------

    def acquire_domain_validations(self, validator, domains):
        ''' requests for all given domains domain validations
            If we have cached a valid challenge return this.
            Expired challenges will clear automatically; invalided challenges
            will not.

            :param domains: List of domains to validate
            :type domains: list of `str`
            :returns: Challenges for the requested domains
            :rtype: acme.messages.ChallengeBody
        '''
        while True:
            authzrs = []
            request_events = []  # we wait for same challenges to be requested
            wait_until = None
            for domain in domains:
                try:
                    authzr = self.authzrs.get(domain, None)
                    if authzr:
                        authzrs.append(self.evaluate_domain_authorization(authzr, validator))
                    else:
                        authzrs.append(self.new_domain_authorization(validator, domain))
                except exceptions.AuthorizationNotYetProcessed as e:
                    # we have to wait until validation is decided
                    wait_until = max(wait_until or datetime.min,
                                     e.wait_until)
                except exceptions.AuthorizationNotYetRequested as e:
                    request_events.append(e.event)
            # some challenges are new; wait until someone (hopefully the ACME
            # server) requests the challenges until further processing
            if request_events:
                for event in request_events:
                    event.wait(timeout=10)
                # give the ACME server 1 second to process the response
                wait_until = max(wait_until or datetime.min,
                                 datetime.now() + timedelta(seconds=1))
            if wait_until:
                wait_time = wait_until - datetime.now()
                self.log('Next authorization poll in {}s ...'.format(wait_time))
                time.sleep(wait_time.total_seconds())
            else:
                return authzrs

    def evaluate_domain_authorization(self, authzr, validator, refresh_timer=None):
        ''' Processes a given AuthorizationResource that was fetch from
            the authzrs cache or updated by `refresh_domain_authorization` /
            `acme.client.Client.poll`.

            Renew revoked or expired ones.
            Refresh pending/processing authorizations

            :param acme.messages.AuthorizationResource authzr: the authzr in
                question
            :return: a valid authzr
            :rtype: acme.messages.AuthorizationResource
            :raises acmems.exceptions.AuthorizationNotYetProcessed: We have to
                wait while the ACME server processes the autzr
            :raises acmems.exceptions.AuthorizationNotYetRequested: new authzr
                created; have to wait until someone requests it
            :raises acmems.exceptions.ChallengesUnknownStatus: unknown status
            :raises acmems.exceptions.NoChallengeMethodsSupported: HTTP01 is
                not supported
            :raises acmems.exceptions.ChallengeFailed: challenge failed
        '''
        authz = authzr.body
        domain = authz.identifier.value
        if authz.status.name == 'valid':
            return authzr
        elif self.invalid_authz(authz):
            # remove auth
            self.authzrs.pop(domain)
            # request new validation
            return self.new_domain_authorization(validator, domain)
        elif authz.status.name == 'invalid':
            message = '; '.join(c.error.detail
                                for c in authz.challenges
                                if c.status.name == 'invalid')
            raise exceptions.ChallengeFailed(domain, message, authzr.uri)
        elif authz.status.name in ('pending', 'processing'):
            # validation in process; check for updates ...
            if refresh_timer:  # we already did a refresh, wait for changes
                raise exceptions.AuthorizationNotYetProcessed(refresh_timer)
            else:
                return self.refresh_domain_authorization(validator, domain)
        else:
            raise exceptions.ChallengesUnknownStatus(authz.status.name)

    def invalid_authz(self, authz):
        if authz.status.name == 'revoked':
            return True
        return (authz.expires.replace(tzinfo=None) - datetime.utcnow()) \
            < timedelta(seconds=2 * 24 * 3600)  # two days

    def refresh_domain_authorization(self, validator, domain):
        ''' Refreshes a authorization for status changes

            :param str domain: domain name for the authorization
            :return: a valid authzr
            :rtype: acme.messages.AuthorizationResource
            :raises acmems.exceptions.AuthorizationNotYetProcessed: We have to
                wait while the ACME server processes the autzr
            :raises acmems.exceptions.AuthorizationNotYetRequested: new authzr
                created; have to wait until someone requests it
            :raises acmems.exceptions.ChallengesUnknownStatus: unknown status
            :raises acmems.exceptions.NoChallengeMethodsSupported: HTTP01 is
                not supported
        '''
        self.log('Refresh authorization for {}'.format(domain))
        self.authzrs[domain], resp = self.client.poll(self.authzrs[domain])
        return self.evaluate_domain_authorization(
            self.authzrs[domain], validator,
            refresh_timer=self.client.retry_after(resp, default=10))

    def new_domain_authorization(self, validator, domain):
        ''' Requests a complete new authorization for the given domain

            :param str domain: domain name for the authorization
            :return: a valid authzr
            :rtype: acme.messages.AuthorizationResource
            :raises acmems.exceptions.AuthorizationNotYetProcessed: We have to
                wait while the ACME server processes the autzr
            :raises acmems.exceptions.AuthorizationNotYetRequested: new authzr
                created; have to wait until someone requests it
            :raises acmems.exceptions.ChallengesUnknownStatus: unknown status
            :raises acmems.exceptions.NoChallengeMethodsSupported: HTTP01 is
                not supported
        '''
        self.log('Requesting new authorization for {}'.format(domain))
        try:
            authzr = self.client.request_domain_challenges(
                domain, self.registration.new_authzr_uri)
            authz = authzr.body
        except acme.messages.Error as e:
            if e.typ == 'urn:acme:error:malformed':
                raise exceptions.InvalidDomainName(domain, e.detail)
            raise
        self.authzrs[domain] = authzr

        if not validator.new_authorization(authz, self.client, self.key, domain):
            # HTTP01 is not support; no clue what to do ...
            raise exceptions.NoChallengeMethodsSupported(
                'No supported challenge methods were offered for {}.'
                .format(domain))

    # cert_response.body and chain now hold OpenSSL.crypto.X509 objects.
    # Convert them to PEM format.
    @staticmethod
    def cert_to_pem(cert):
        return OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, cert).decode('utf-8')

    def issue_certificate(self, csr, authzrs):
        # Request a certificate using the CSR and some number of domain validation challenges.
        self.log("Requesting a certificate.")
        try:
            cert_response = self.client.request_issuance(acme.jose.ComparableX509(csr), authzrs)
        except acme.messages.Error as e:
            if e.typ == "urn:acme:error:rateLimited":
                raise exceptions.RateLimited(e.detail)
            raise  # unhandled

        certs = [self.cert_to_pem(cert_response.body)]

        # Get the certificate chain.
        for cert in self.client.fetch_chain(cert_response):
            certs.append(self.cert_to_pem(cert))

        return tuple(certs)



import warnings
from fnmatch import fnmatch
import hmac
import hashlib

from OpenSSL import crypto
import pyasn1.type
from pyasn1.codec.der import decoder
from ndg.httpsclient.subj_alt_name import SubjectAltName as BaseSubjectAltName
from IPy import IP

from acmems import exceptions


### Note: This is a slightly bug-fixed version of same from ndg-httpsclient.
class SubjectAltName(BaseSubjectAltName):
    '''ASN.1 implementation for subjectAltNames support'''

    # There is no limit to how many SAN certificates a certificate may have,
    #   however this needs to have some limit so we'll set an arbitrarily high
    #   limit.
    sizeSpec = pyasn1.type.univ.SequenceOf.sizeSpec + \
        pyasn1.type.constraint.ValueSizeConstraint(1, 1024)


class Authenticator():
    def __init__(self, config=None):
        self.config = config
        self.blocks = []

    def parse_block(self, name, options):
        self.blocks.append(Block(name, options, self.config))

    def process(self, client_address, headers, rfile):
        return Processor(self, client_address, headers, rfile)


class IPAuthMethod():
    ''' Autentication by source IP'''
    option_names = ['ip']

    def __init__(self, ips=None):
        self.ips = ips or []

    def parse(self, option, value):
        assert option == 'ip'
        self.ips.append(IP(value))

    def possible(self, processor):
        return self.check(processor)

    def check(self, processor):
        cip = IP(processor.client_address[0])
        for ip in self.ips:
            if cip in ip:
                return True
        return False


class HmacAuthMethod():
    ''' Authentication by HMAC / secret key '''
    option_names = ['hmac_type', 'hmac_key']

    def parse(self, option, value):
        if option == 'hmac_type':
            self.name = value
            self.hmac = getattr(hashlib, self.name)
        elif option == 'hmac_key':
            self.key = value.encode('utf-8')

    def parse_authentification_header(self, processor):
        if 'Authentication' not in processor.headers:
            return (None, {})
        else:
            name, opts = processor.headers['Authentication'].split(' ', 1)
            opts = {opt.split('=')[0]: opt.split('=', 1)[1] for opt in opts.split(', ')}
            return name, opts

    def possible(self, processor):
        name, opts = self.parse_authentification_header(processor)
        if name != 'hmac':
            return False
        if opts.get('name', None) != self.name:
            return False
        if 'hash' not in opts:
            return False
        return True

    def check(self, processor):
        name, opts = self.parse_authentification_header(processor)
        csrhash = hmac.new(self.key, processor.csrpem, digestmod=self.hmac).hexdigest()
        if hmac.compare_digest(csrhash, opts['hash']):
            return True
        else:
            return False


class AllAuthMethod():
    ''' Allow all authentication'''
    option_names = ['all']

    def parse(self, option, value):
        assert option == 'all'
        assert value == 'yes'

    def possible(self, processor):
        return True

    def check(self, processor):
        return True


class Block():
    ''' One authentication block - combination of authentications
        and list of allowed domains
    '''
    def __init__(self, name, options, config):
        self.name = name
        self.methods = []
        self.domain_matchers = []
        self.validator = None
        self.storage = None
        self.parse(options, config)

    def possible(self, processor):
        if not self.domain_matchers:
            return False
        for method in self.methods:
            if not method.possible(processor):
                print('block {} excluded by {}'.format(self.name, method.__class__.__name__))
                return False
        return True

    def check(self, processor):
        for method in self.methods:
            if not method.check(processor):
                return False
        # check matching of domain names
        for dns_name in processor.dns_names:
            for matcher in self.domain_matchers:
                if fnmatch(dns_name, matcher):
                    break
            else:
                return False
        return True

    def parse(self, options, config):
        unused_methods = [IPAuthMethod, AllAuthMethod, HmacAuthMethod]
        for option, value in options:
            if option == 'domain':
                self.domain_matchers.append(value)
                continue
            if option == 'verification':
                try:
                    self.validator = config.validators[value.strip()]
                except KeyError:
                    from acmems.config import UnknownVerificationError
                    raise UnknownVerificationError('Validator "{}" undefined'.format(value.strip()))
                continue
            if option == 'storage':
                try:
                    self.storage = config.storages[value.strip()]
                except KeyError:
                    from acmems.config import UnknownStorageError
                    raise UnknownStorageError('Storage "{}" undefined'.format(value.strip()))
                continue
            for method in self.methods:
                if option in method.option_names:
                    method.parse(option, value)
                    break
            else:  # no known method processes this option
                for method in unused_methods:
                    if option in method.option_names:
                        break
                else:
                    from acmems.config import UnusedOptionWarning
                    warnings.warn('Option unknown [auth "{}"]{} = {}'.format(self.name, option, value),
                                  UnusedOptionWarning, stacklevel=2)
                    break
                unused_methods.remove(method)
                self.methods.append(method())
                self.methods[-1].parse(option, value)
        if self.validator is None:
            self.validator = config.default_validator
            if self.validator is False:
                from acmems.config import UnknownVerificationError
                raise UnknownVerificationError('auth "{}" does not define a validator and the default one is disabled'.format(self.name))
        if self.storage is None:
            self.storage = config.default_storage
            if self.storage is False:
                from acmems.config import UnknownStorageError
                raise UnknownStorageError('auth "{}" does not define a storage and the default one is disabled'.format(self.name))


class Processor():
    ''' Helper object to process a request, check authentication,
        reads and parse CSR
    '''
    def __init__(self, auth, client_address, headers, rfile):
        self.auth = auth
        self.client_address = client_address
        self.headers = headers
        self.rfile = rfile

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def acceptable(self):
        ''' process the given request parameter for a CSR signing request and
            decide whether this request is allowed or not.

            :param client_ip str: The source IP of the client (TCP level)
            :param dict headers: The request header
            :param callable get_body: function to read in body (CSR)
            :return bool: whether request should be accepted
        '''
        self.validator = None
        self.storage = None
        # 1. precheck
        possible_blocks = []
        for block in self.auth.blocks:
            if block.possible(self):
                possible_blocks.append(block)
        if not possible_blocks:
            return False
        # 2. process CSR
        try:
            self.read_and_parse_csr()
        except crypto.Error:
            raise exceptions.PayloadInvalid()
        self.accepted_block = None
        # 3. final check
        for block in possible_blocks:
            if block.check(self):
                self.validator = block.validator
                self.storage = block.storage
                return True
        return False

    def read_and_parse_csr(self):
        content_length = int(self.headers['Content-Length'])
        if self.auth.config and content_length > self.auth.config.max_size:
            raise exceptions.PayloadToLarge(size=content_length, allowed=self.auth.config.max_size)
        self.csrpem = self.rfile.read(content_length)
        self.csr = crypto.load_certificate_request(crypto.FILETYPE_PEM, self.csrpem)
        self.common_name = self.csr.get_subject().CN
        self.dns_names = []
        for ext in self.csr.get_extensions():
            if ext.get_short_name() != b'subjectAltName':
                continue
            general_names = SubjectAltName()
            data = ext.get_data()
            decoded_dat = decoder.decode(data, asn1Spec=general_names)
            for name in decoded_dat:
                if not isinstance(name, SubjectAltName):
                    continue
                for entry in range(len(name)):
                    component = name.getComponentByPosition(entry)
                    if component.getName() != 'dNSName':
                        continue
                    self.dns_names.append(str(component.getComponent()))
        if self.common_name not in self.dns_names:
            self.dns_names.insert(0, self.common_name)



class AcmeException(Exception):
    ''' Base exception call to be able to catch all ACMEMS specific
        errors
    '''
    pass


class NoChallengeMethodsSupported(AcmeException):
    ''' The domain can not be validated HTTP01
    '''
    pass


class ChallengeFailed(AcmeException):
    ''' The challenge to validate the requested domain failed.

        :ivar str domain: the domain which the challenge should validate
        :ivar str message: message description from ACME server
        :ivar str challenge_uri: the URI of the failed challenge
    '''
    def __init__(self, domain, message, challenge_uri):
        self.domain = domain
        self.message = message
        self.challenge_uri = challenge_uri

    def __str__(self):
        return "The challenge for {} failed: {}.".format(self.domain, self.message)


class ChallengesUnknownStatus(AcmeException):
    ''' We do not known the status of the challenge. No clue what to do '''
    pass


class AuthorizationNotYetProcessed(AcmeException):
    ''' The authorization is be processed; until the next refresh it should
        at least be wait until `~wait_until`

        :ivar datetime.datetime wait_until: first allowed retry time
    '''
    def __init__(self, wait_until):
        self.wait_until = wait_until


class AuthorizationNotYetRequested(AcmeException):
    ''' The newly created authorization challenge, was installed, but has not
        yet been requested by any client and is therefore currently pending
        or invalid.

        :ivar threading.Event event: event that will be signaled if someone
            requests the challenge.
    '''
    def __init__(self, event):
        self.event = event


class RateLimited(AcmeException):
    ''' To many requests '''
    pass


class AccountError(AcmeException):
    ''' Generic account error - e.g.
        - could not read private key
        - could not refresh the registration
    '''
    pass


class NeedToAgreeToTOS(AccountError):
    ''' We are registered at the ACME server. But to use it,
        we need to accept the "Terms of Service"
    '''
    def __init__(self, url):
        self.url = url


class InvalidDomainName(AcmeException):
    ''' The domain name is not excepted by the ACME server.

        :ivar str domain: the domain that was rejected
        :ivar str detail: the reject reason as string
    '''
    def __init__(self, domain, detail):
        self.domain = domain
        self.detail = detail

    def __str__(self):
        return '{} is not a domain name that the ACME server can issue ' \
            'a certificate for: {}'.format(self.domain, self.detail)


class PayloadToLarge(AcmeException):
    ''' The payload (CSR) it to large

        :ivar int size: the request size to upload (in bytes)
        :ivar int allowed: the maximal size in bytes
    '''
    def __init__(self, size, allowed):
        self.size = size
        self.allowed = allowed


class PayloadInvalid(AcmeException):
    ''' The payload is not a valid CSR '''
    pass



from setuptools import setup  # Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

from acmems import version

setup(
    name='acme-mgmtserver',
    version=version.STRING,

    description='Basic Python Server to execute ACME instead of dump clients',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/mswart/acme-mgmtserver',

    # Author details
    author='Malte Swart',
    author_email='mswart@devtation.de',

    # Choose your license
    license='GPL',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: System Administrators',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Security :: Cryptography',
        'Topic :: System :: Networking',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='acme server client proxy letsencrypt',

    packages=['acmems'],

    install_requires=[
        'acme>=0.2.0',
        'ndg-httpsclient',  # urllib3 InsecurePlatformWarning (#304)
        'pyasn1',  # urllib3 InsecurePlatformWarning (#304)
        'PyOpenSSL>=0.15',  # X509Req.get_extensions (>=0.15)
        'IPy',
    ],

    scripts=['bin/acmems', 'bin/acme-register'],
)



import socket
from threading import Thread, Event
from datetime import datetime, timedelta
import json
import urllib.request
import warnings

import acme.client

from acmems.config import ConfigurationError, UnusedOptionWarning
from acmems.server import ThreadedACMEServerByType, ACMEHTTPHandler
from acmems import exceptions


class ChallengeImplementor():
    def __init__(self, type, name, options):
        self.type = type
        self.name = name
        self.parse(options)


class HttpChallengeImplementor(ChallengeImplementor):
    def parse(self, options):
        self.listeners = None
        for option, value in options:
            if option == 'listener':
                if self.listeners is None:
                    self.listeners = []
                if value == '':  # disable listener
                    continue
                if ':' not in value:
                    raise ConfigurationError('unix socket are currenlty not supported as listeners')
                host, port = value.rsplit(':', 1)
                if host[0] == '[' and host[-1] == ']':
                    host = host[1:-1]
                self.listeners += socket.getaddrinfo(host, int(port), proto=socket.IPPROTO_TCP)
        if self.listeners is None:
            self.listeners = socket.getaddrinfo('0.0.0.0', 1380, proto=socket.IPPROTO_TCP) \
                + socket.getaddrinfo('::', 1380, proto=socket.IPPROTO_TCP)

    def start(self):
        services = []
        bound_handler = lambda *args, **kwargs: ACMEHTTPHandler(self, *args, **kwargs)
        for http_listen in self.listeners:
            http_service = ThreadedACMEServerByType[http_listen[0]](http_listen[4], bound_handler)
            thread = Thread(target=http_service.serve_forever,
                            daemon=True,
                            name='http service to server validation request')
            thread.start()
            services.append((http_service, thread))
        self.responses = {}
        return services

    def new_authorization(self, authz, client, key, domain):
        for combination in authz.combinations:
            if len(combination) == 1:
                challenger = authz.challenges[combination[0]]
                challenge = challenger.chall
                if isinstance(challenge, acme.challenges.HTTP01):
                    # store (and deliver) needed response for challenge
                    content = challenge.validation(key)
                    event = Event()
                    self.responses.setdefault(domain, {})
                    self.responses[domain][challenge.path] = (content, event)

                    # answer challenges / give ACME server go to check challenge
                    resp = challenge.response(key)
                    client.answer_challenge(challenger, resp)

                    # we can wait until this challenge is first requested ...
                    raise exceptions.AuthorizationNotYetRequested(event)
        else:
            return False

    def response_for(self, host, path):
        ''' request a response for a given request

        :param str host: Hostname of the request
        :param str path: Requested path (e.g. /.well-known/acme-challenges/?)
        :raises KeyError: Unknown host or path; return 404
        '''
        content, event = self.responses[host][path]
        event.set()
        return content


class DnsChallengeImplementor(ChallengeImplementor):
    """ WIP """
    def start(self):
        pass

    def new_authorization(self, authz, client, key, domain):
        for combination in authz.combinations:
            if len(combination) == 1:
                challenger = authz.challenges[combination[0]]
                challenge = challenger.chall
                if isinstance(challenge, acme.challenges.DNS01):
                    response, validation = challenge.response_and_validation(key)

                    self.add_entry(challenge.validation_domain_name(domain) + '.', validation)

                    # answer challenges / give ACME server go to check challenge
                    client.answer_challenge(challenger, response)

                    # we can wait until this challenge is first requested ...
                    raise exceptions.AuthorizationNotYetProcessed(datetime.now() + timedelta(seconds=2))
        else:
            return False


class DnsChallengeBoulderImplementor(DnsChallengeImplementor):
    """ WIP """
    def parse(self, options):
        self.set_txt_url = None
        for option, value in options:
            if option == 'set-txt_url':
                self.set_txt_url = value
        if self.set_txt_url is None:
            self.set_txt_url = 'http://localhost:8055/set-txt'

    def add_entry(self, entry, value):
        task = json.dumps({
            'host': entry,
            'value': value
        }).encode('utf-8')

        response = urllib.request.urlopen(self.set_txt_url, task)
        assert response.code is 200


class DnsChallengeDnsUpdateImplementor(DnsChallengeImplementor):
    """ WIP """
    def parse(self, options):
        self.dns_servers = None
        self.ttl = None
        self.timeout = None
        for option, value in options:
            if option == 'dns-server':
                if self.dns_servers is None:
                    self.dns_servers = []
                self.dns_servers.append(value)
            elif option == 'ttl':
                self.ttl = int(value)
            elif option == 'timeout':
                self.timeout = int(value)
            else:
                warnings.warn('Option unknown [verification "{}"]{} = {}'.format(self.name, option, value),
                              UnusedOptionWarning, stacklevel=2)
        if self.dns_servers is None:
            self.dns_servers = ['127.0.0.1']
        if self.ttl is None:
            self.ttl = 60
        if self.timeout is None:
            self.timeout = 5

    def add_entry(self, entry, value):
        import dns
        import dns.update
        import dns.query
        upd = dns.update.Update(self.select_zone(entry),
                                #keyring=dns.tsigkeyring.from_text({keyname: key}),
                                #keyalgorithm=algo)
                                )
        upd.add(entry, self.ttl, 'TXT', value)
        try:
            response = dns.query.tcp(upd, self.dns_servers[0], timeout=self.timeout)
            rcode = response.rcode()
            if rcode != dns.rcode.NOERROR:
                rcode_text = dns.rcode.to_text(rcode)
                raise ValueError(rcode_text)
            return response
        except Exception as e:
            raise ValueError('could not update {}: {}'.format(e.__class__.__name__, e))

    def select_zone(self, entry):
        parts = entry.split('.')
        return '.'.join(parts[-3:])


implementors = {
    'http01': HttpChallengeImplementor,
    'dns01-boulder': DnsChallengeBoulderImplementor,
    'dns01-dnsUpdate': DnsChallengeDnsUpdateImplementor,
}


def setup(type, name, options):
    try:
        return implementors[type](type, name, options)
    except KeyError:
        raise ConfigurationError('Unsupported challenge type "{}"'.format(type))
